<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 3/23/16
 * Time: 1:31 PM
 */

namespace qfq;

require_once(__DIR__ . '/Store.php');
require_once(__DIR__ . '/../Database.php');
require_once(__DIR__ . '/../Constants.php');
require_once(__DIR__ . '/../helper/HelperFormElement.php');
require_once(__DIR__ . '/../exceptions/UserFormException.php');

class FillStoreForm {

    /**
     * @var Store
     */
    private $store = null;

    /**
     * @var Database
     */
    private $db = null;

    /**
     * @var array
     */
    private $feSpecNative = array();

    /**
     *
     */
    public function __construct() {
        $this->store = Store::getInstance();
        $this->db = new Database();
        $this->feSpecNative = $this->loadFormElementsBasedOnSIP();
    }

    /**
     * Loads a minimal definition of FormElement of the form specified in SIP.
     *
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    private function loadFormElementsBasedOnSIP() {
        $formName = $this->store->getVar(SIP_FORM, STORE_SIP);

        // Preparation for Log, Debug
        $this->store->setVar(SYSTEM_FORM, $formName, STORE_SYSTEM);

        $feSpecNative = $this->db->sql(SQL_FORM_ELEMENT_SIMPLE_ALL_CONTAINER, ROW_REGULAR, [$formName]);

        if (count($feSpecNative) === 0) {
            throw new UserFormException('Form not found or multiple forms with the same name.', ERROR_FORM_NOT_FOUND);
        }

        HelperFormElement::explodeParameterInArrayElements($feSpecNative);

        return $feSpecNative;
    }

    /**
     * Copies all current form parameter from STORE_CLIENT to STORE_FORM. Checks the values against FormElement
     * definition and throws an exception if check fails. FormElements.type=hidden will be taken from STORE_SIP.
     *
     * @throws CodeException
     * @throws UserFormException
     */
    public function process() {
        $html = '';
        $newValues = array();

        $clientValues = $this->store->getStore(STORE_CLIENT);

        // Retrieve SIP vars, e.g. for HIDDEN elements.
        $sipValues = $this->store->getStore(STORE_SIP);

        // Copy SIP Values; not necessarily defines as FormElements.
        foreach ($sipValues as $key => $value) {
            switch ($key) {
                case SIP_SIP:
                case SIP_RECORD_ID:
                case SIP_FORM:
                case SIP_TABLE:
                case SIP_URLPARAM:
                case 'id':
                    break;
                default:
                    $newValues[$key] = $value;
                    break;
            }
        }

        // Iterate over all formelements. Sanatize values. Built an assoc array $newValues.
        foreach ($this->feSpecNative AS $formElement) {

            // Never get a predefined 'id'
            if ($formElement['name'] === 'id')
                continue;

            // Preparation for Log, Debug
            $this->store->setVar(SYSTEM_FORM_ELEMENT, Logger::formatFormElementName($formElement), STORE_SYSTEM);

            // evaluate current FormElement: e.g. FE_MODE_SQL
            $evaluate = new Evaluate($this->store, $this->db);
            $formElement = $evaluate->parseArray($formElement, $debugStack);

            // Get related formElement.
            // construct the field name used in the form
            $clientFieldName = HelperFormElement::buildFormElementId($formElement['name'], $sipValues[SIP_RECORD_ID]);

            // Some Defaults
            $formElement = Support::setFeDefaults($formElement);

            // Preparation for Log, Debug
//            $this->store->setVar(SYSTEM_FORM_ELEMENT, $formElement['name'] . ' / ' . $formElement['id'], STORE_SYSTEM);

            if ($formElement[FE_TYPE] == 'hidden') {
                // Hidden elements will be transferred by SIP
                if (!isset($sipValues[$formElement['name']])) {
                    throw new CodeException("Missing the hidden field '" . $formElement['name'] . "' in SIP.", ERROR_MISSING_HIDDEN_FIELD_IN_SIP);
                }

                $newValues[$formElement['name']] = $sipValues[$formElement['name']];
                continue;
            }

            if ($formElement[FE_MODE] === FE_MODE_REQUIRED) {
                if (!isset($clientValues[$clientFieldName]) || ($clientValues[$clientFieldName] === '')) {
                    throw new UserFormException("Missing required value.", ERROR_REQUIRED_VALUE_EMPTY);
                }
            }

            switch ($formElement[FE_MODE]) {
                case FE_MODE_REQUIRED:
                case FE_MODE_SHOW:
                    if (isset($clientValues[$clientFieldName])) {

                        // SELECT MULTI or CHECKBOX MULTI: delivered as array - implode them.
                        if (is_array($clientValues[$clientFieldName])) {
                            // E.g. Checkboxes needs a 'HIDDEN' HTML input to detect 'unset' of values. These 'HIDDEN' element
                            //  needs to be removed, if there is at least one checkbox is checked (=submitted)
                            if (count($clientValues[$clientFieldName]) > 1)
                                array_shift($clientValues[$clientFieldName]);

                            $clientValues[$clientFieldName] = implode(',', $clientValues[$clientFieldName]);
                        }

                        switch ($formElement[FE_TYPE]) {
                            case 'date':
                            case 'datetime':
                            case 'time':
                                if ($clientValues[$clientFieldName] !== '') // do not check empty values
                                    $newValues[$formElement['name']] = $this->doDateTime($formElement, $clientValues[$clientFieldName]);
                                break;
                            default:
                                $newValues[$formElement['name']] = Sanitize::sanitize($clientValues[$clientFieldName],
                                    $formElement['checkType'], $formElement['checkPattern'], SANATIZE_EXCEPTION);
                                break;
                        }
                    }
                    break;

                case FE_MODE_READONLY:
                case FE_MODE_HIDDEN:
                    continue;
                default:
                    throw new CodeException("Unknown mode: " . $formElement[FE_MODE], ERROR_UNKNOWN_MODE);
            }
        }

        $this->store->setVarArray($newValues, STORE_FORM, true);
    }

    /**
     * Check  $value as date/datime/time value and convert it to FORMAT_DATE_INTERNATIONAL.
     *
     * @param array $formElement - if not set, set $formElement['dateFormat']
     * @param string $value - date/datetime/time value in format FORMAT_DATE_INTERNATIONAL or FORMAT_DATE_GERMAN
     * @return string - checked datetime string
     * @throws UserFormException
     */
    private function doDateTime(array &$formElement, $value) {

        $regexp = Support::dateTimeRegexp($formElement[FE_TYPE], $formElement['dateFormat']);

        if (1 !== preg_match('/' . $regexp . '/', $value, $matches)) {
            $placeholder = Support::getDateTimePlaceholder($formElement);
            throw new UserFormException("DateTime format not recognized: $placeholder / $value ", ERROR_DATE_TIME_FORMAT_NOT_RECOGNISED);
        }

        $showTime = $formElement[FE_TYPE] == 'date' ? '0' : '1';
        $value = Support::convertDateTime($value, FORMAT_DATE_INTERNATIONAL, '1', $showTime, $formElement[FE_SHOW_SECONDS]);

        return $value;
    }
}